/* 
 * File:   Dirichlet.cpp
 * Author: wicker
 * 
 * Created on 8 septembre 2011, 09:18
 */

#include <iostream>
#include <cfloat>
#include <cmath>
#include "Lawc.h"
#include "Dirichlet.h"
#include "matrix.h"
#include "PointsSet.h"
#define MATHLIB_STANDALONE
#include <Rmath.h>
using namespace std;

Dirichlet::Dirichlet() {
}

Dirichlet::Dirichlet(int dimension, bool diag) {
    this->dimension = dimension;
    this->diag = false;

    this->alphas = new double[this->dimension];
}

Dirichlet::~Dirichlet() {
}

void Dirichlet::estimation(int nbPoints, double **values) {
    /*declaration de variables*/
    int i, j, k, nbIterations, nbIterationsMax, probleme, unGradientNegatif, unAlphaInferieurAUn;
    double somme, sommeAlphas, *variances, difference, *gradient, normeDeriveeCarre;
    double *nouveauxAlphas, pas, logVraisemblanceCourante, nouvelleLogVraisemblance, *pasMax;
    double **inverseHessien, denominateur, a, alpha0Max, alpha0MaxMoins, alpha0MaxPlus, *fj;
    double differenceAlphas, sommeInverses;

    /*pour tests*/
    double premiersAlphas[50];
    /*fin declaration de variables*/

    /*allocation memoire*/
    variances = new double[this->dimension];
    gradient = new double[this->dimension];
    inverseHessien = new double*[this->dimension];
    for (i = 0; i < this->dimension; i++) {
        inverseHessien[i] = new double[this->dimension];
    }
    nouveauxAlphas = new double[this->dimension];
    fj = new double[this->dimension];
    /*  cluster.alphas=(double *)malloc(sizeof(double)*this->dimension);*/

    /*fin allocation memoire*/
    gradient[0] = 5.0;
    nbIterationsMax = 100;

    /*on calcule les fj, estimation par les moments*/
    sommeInverses = 0;
    for (k = 0; k < this->dimension; k++) {
        fj[k] = 0;
        for (j = 0; j < nbPoints; j++) {
            fj[k] += values[k][j];
        }
        fj[k] /= (double) nbPoints;
        sommeInverses += 1.0 / fj[k];
    }

    /*on calcule a en resolvant une equation du second degre apres approximation de la logvraisemblance*/
    a = 0;
    for (k = 0; k < this->dimension; k++) {
        somme = 0.0;
        for (j = 0; j < nbPoints; j++) {
            somme += log(values[k][j]);
        }
        a += fj[k]*(somme - (double) nbPoints * log(fj[k]));
    }

    alpha0Max = -(double) (nbPoints * (this->dimension - 1))*0.5 -
            sqrt((double) nbPoints * ((double) nbPoints / 4.0 *
            (double) (this->dimension - 1)*(double) (this->dimension - 1) -
            a * (sommeInverses - 1.0) / 3.0));
    alpha0Max /= 2.0 * a;
    sommeAlphas = 0;
    for (j = 0; j < this->dimension; j++) {
        this->alphas[j] = fj[j] * alpha0Max;
        sommeAlphas += this->alphas[j];
        premiersAlphas[j] = this->alphas[j];
        if (!((premiersAlphas[j] > DBL_MIN) && (premiersAlphas[j] < DBL_MAX))) {
            cout << premiersAlphas[j] << " " << alpha0Max << " " << a << " " << somme << endl;
            break;
        }
    }

    /*calcul de la log-vraisemblance associee a ces parametres*/
    logVraisemblanceCourante = this->computeLogDensityForNPoints(nbPoints, values);

    /*on calcule les parametres par la methode du gradient a pas variable*/
    nbIterations = 0;
    while (1) {
        normeDeriveeCarre = 0;

        /*on calcule le gradient*/
        for (j = 0; j < this->dimension; j++) {
            gradient[j] = (digamma(sommeAlphas) - digamma(this->alphas[j]))*(double) nbPoints;

            for (k = 0; k < nbPoints; k++) {
                gradient[j] += log(values[j][k]);
            }
            normeDeriveeCarre += gradient[j] * gradient[j];
        }

        denominateur = 1.0 / ((double) nbPoints * trigamma(sommeAlphas));
        for (j = 0; j < this->dimension; j++) {
            denominateur -= 1.0 / ((double) nbPoints * trigamma(this->alphas[j]));
        }

        for (j = 0; j < this->dimension; j++) {
            for (k = j; k < this->dimension; k++) {
                inverseHessien[j][k] = 1.0 / (-(double) nbPoints * trigamma(this->alphas[j])*
                        (double) nbPoints * trigamma(this->alphas[k])) / denominateur;
                inverseHessien[k][j] = inverseHessien[j][k];
            }
            inverseHessien[j][j] -= 1.0 / ((double) nbPoints * trigamma(this->alphas[j]));
        }

        for (j = 0; j < this->dimension; j++) {
            nouveauxAlphas[j] = this->alphas[j];
            for (k = 0; k < this->dimension; k++) {
                nouveauxAlphas[j] -= inverseHessien[j][k] * gradient[k];
            }
        }

        /*calcul de la log-vraisemblance associee aux nouveaux parametres*/
        nouvelleLogVraisemblance = this->computeLogDensityForNPoints(nbPoints, values);

        differenceAlphas = 0;
        for (j = 0; j < this->dimension; j++) {
            differenceAlphas += (nouveauxAlphas[j] - this->alphas[j])*
                    (nouveauxAlphas[j] - this->alphas[j]);
        }
        differenceAlphas = sqrt(differenceAlphas);

        sommeAlphas = 0;
        for (j = 0; j < this->dimension; j++) {
            if (!((this->alphas[j] > DBL_MIN) && (this->alphas[j] < 999999999.0))) {
                break;
            }
        }

        for (j = 0; j < this->dimension; j++) {
            this->alphas[j] = nouveauxAlphas[j];
            sommeAlphas += this->alphas[j];
        }

        if (fabs(logVraisemblanceCourante - nouvelleLogVraisemblance) < DBL_MIN) {
            break;
        }
        logVraisemblanceCourante = nouvelleLogVraisemblance;
        nbIterations++;
        if (nbIterations > nbIterationsMax) {
            break;
        }
    }


    /*desallocation memoire*/
    delete []variances;
    delete []gradient;
    delete []nouveauxAlphas;
    for (i = 0; i < this->dimension; i++) {
        delete []inverseHessien[i];
    }
    delete []inverseHessien;
    delete []fj;
    /*fin desallocation memoire*/

}


/************************************************************************************************/
/*                                                                                              */
/*Procedure de calcul de la log-vraisemblance d'une classe d'individus pour une loi de Dirichlet*/
/*                                                                                              */

/************************************************************************************************/

double Dirichlet::computeLogDensityForNPoints(int nbPoints, double **values) {
    /*declaration de variables*/
    int i, j;
    double somme, logDensity;
    /*fin declaration de variables*/

    logDensity = 0;
    somme = 0;
    for (i = 0; i<this->dimension; i++) {
        somme += this->alphas[i];
    }
    logDensity += lgammafn(somme);
    for (i = 0; i<this->dimension; i++) {
        logDensity -= lgammafn(this->alphas[i]);
    }
    logDensity *= (double) nbPoints;

    for (j = 0; j<this->dimension; j++) {
        somme = 0;
        for (i = 0; i < nbPoints; i++) {
            somme += log(values[j][i]);
        }
        logDensity += somme * (alphas[j] - 1.0);
    }

    return logDensity;
}

/***********************************/
/*compute log density for one point*/

/***********************************/

double Dirichlet::computeLogDensity(double *values) {
    /*declaration de variables*/
    int i, j, k;
    double logDensity, somme, commun;
    /*fin declaration de variables*/

    commun = 0;
    somme = 0;
    for (j = 0; j < this->dimension; j++) {
        somme += this->alphas[j];
        commun -= lgammafn(this->alphas[j]);
    }
    commun += lgammafn(somme);
    logDensity = commun;
    for (k = 0; k < this->dimension; k++) {
        logDensity += (this->alphas[k] - 1.0) * log(values[k]);
    }

    return logDensity;
}


double Dirichlet::computeLogDensityNoise(int nbPoints,double **values) {
  /*declaration de variables*/
  double logDensity,volumeSimplex;
  /*fin declaration de variables*/

  /*calcul du volume d'un tetraedre*/
  calculVolumeSimplex(this->dimension-1,&volumeSimplex);

  logDensity=-log(volumeSimplex);

  return logDensity;
}

void Dirichlet::print() {
    cout << "alphas" << endl;
    for (int i = 0; i<this->dimension; i++) {
        cout << this->alphas[i] << endl;
    }


}

int Dirichlet::complexity() {
    int temp;
    /*a completer*/

    temp = this->dimension;
    return temp;
}


