/* 
 * File:   FisherVonMises.cpp
 * Author: wicker
 * 
 * Created on 8 septembre 2011, 09:18
 */

#include <iostream>
#include <cfloat>
#include <cmath>
#include "Lawc.h"
#include "FisherVonMises.h"
#include "matrix.h"
#include "PointsSet.h"
#define MATHLIB_STANDALONE
#include <Rmath.h>
using namespace std;

FisherVonMises::FisherVonMises() {
}

FisherVonMises::FisherVonMises(int dimension, bool diag) {

    this->dimension = dimension;
    this->diag = false;

    this->mu = new double[this->dimension];

}

FisherVonMises::~FisherVonMises() {
}

void FisherVonMises::estimation(int nbPoints, double **values) {
    /*declaration de variables*/
    int i, j, k;
    double kappa, Rbar;
    double *vecteurSomme;
    /*fin declaration de variables*/

    /*allocation memoire*/
    vecteurSomme = new double[this->dimension];
    /*fin allocation memoire*/

    /*calcul du mu moyen*/
    calculMuMoyenSpherique(nbPoints, values, this->dimension, this->mu);

    /*calcul de kappa (note nu)*/
    for (i = 0; i < this->dimension; i++) {
        vecteurSomme[i] = 0;
    }
    for (i = 0; i < nbPoints; i++) {
        for (j = 0; j < this->dimension; j++) {
            vecteurSomme[j] += values[j][i];
        }
    }
    Rbar = 0;
    for (i = 0; i < this->dimension; i++) {
        vecteurSomme[i] /= (double) nbPoints;
        Rbar += vecteurSomme[i] * vecteurSomme[i];
    }
    Rbar = sqrt(Rbar);

    kappa = Rbar * ((double) this->dimension - Rbar * Rbar) / (1 - Rbar * Rbar);
    this->nu = kappa;

    /*desallocation memoire*/
    delete []vecteurSomme;
    /*fin desallocation memoire*/

}

/*compute log density for one point*/

double FisherVonMises::computeLogDensity(double *values) {
    /*declaration de variables*/
    int k;
    double logDensity,norme,commun;
    /*fin declaration de variables*/

    commun = ((double) this->dimension / 2.0 - 1.0) * log(this->nu)-((double) this->dimension / 2.0) * log(2.0 * M_PI) -
            log(bessel_j(this->nu, (double) this->dimension / 2.0 - 1.0));
    (double) (this->dimension - 1)*0.5 * log(2.0 * M_PI);

    norme = 0;
    for (k = 0; k < this->dimension; k++) {
        norme += values[k] * this->mu[k];
    }
    logDensity= commun + this->nu*norme;

    return logDensity;
}

double FisherVonMises::computeLogDensityNoise(int nbPoints, double **values) {
    double surfaceSphere, logDensity;
    calculSurfaceSphere(this->dimension - 1, &surfaceSphere);

    logDensity = -log(surfaceSphere);
    return logDensity;
}

void FisherVonMises::print() {
    cout << "nu " << this->nu << endl;

    cout << "mu" << endl;
    for (int i = 0; i<this->dimension; i++) {
        cout << this->mu[i] << endl;
    }


}

int FisherVonMises::complexity() {
    int temp;
    /*a completer*/
    temp = this->dimension;
    return temp;
}
