/* 
 * File:   gaussian.cpp
 * Author: wicker
 * 
 * Created on 8 juillet 2011, 10:39
 */
#include <iostream>
#include "Lawc.h"
#include "Gaussian.h"
#include "matrix.h"
#define MATHLIB_STANDALONE
#include <Rmath.h>
using namespace std;

Gaussian::Gaussian() {

}

Gaussian::Gaussian(int dimension, bool diag) {

    this->dimension = dimension;
    this->diag = diag;

    this->mu = new double[this->dimension];
    if (this->diag == true) {
        this->iSigmaDiag = new double[this->dimension];
    } else {
        this->iSigma = new double *[this->dimension];
        for (int i = 0; i<this->dimension; i++) {
            this->iSigma[i] = new double[this->dimension];
        }
    }
}

Gaussian::~Gaussian() {
}

void Gaussian::estimation(int nbPoints, double **values) {
    int i, j, k;

    /*estimation de la moyenne*/
    for (i = 0; i<this->dimension; i++) {
        this->mu[i] = 0;
        for (j = 0; j < nbPoints; j++) {
            this->mu[i] += values[i][j];
        }
        this->mu[i] /= (double) nbPoints;

    }

    /*estimation de la matrice de variance-covariance*/
    double temp;
    if (diag == true) {
        this->logDeterminant = 0;
        for (i = 0; i<this->dimension; i++) {
            this->iSigmaDiag[i] = 0;
            for (j = 0; j < nbPoints; j++) {
                temp = values[i][j] - this->mu[i];
                this->iSigmaDiag[i] += temp*temp;
            }
            this->iSigmaDiag[i] /= (double) nbPoints;
            this->logDeterminant += log(this->iSigmaDiag[i]);
            this->iSigmaDiag[i] = 1.0 / this->iSigmaDiag[i];
        }
    } else {
        for (i = 0; i<this->dimension; i++) {
            for (j = i; j<this->dimension; j++) {
                this->iSigma[i][j] = 0;
                for (k = 0; k < nbPoints; k++) {
                    this->iSigma[i][j] += (values[i][k] - this->mu[i])*(values[j][k] - this->mu[j]);
                }
                this->iSigma[i][j] /= (double) nbPoints;
                this->iSigma[j][i] = this->iSigma[i][j];
            }
        }
        this->logDeterminant = inversionMatrix(this->dimension, this->iSigma);
    }
}

double Gaussian::computeLogDensity(double *values) {
    double logDensity, norm, common;
    double *differences = new double[this->dimension];

    common = -this->logDeterminant * 0.5 - (double) this->dimension * 0.5 * log(2.0 * M_PI);
    for (int i = 0; i<this->dimension; i++) {
        differences[i] = values[i] - this->mu[i];
    }

    if (this->diag == false) {
        norm = 0;
        for (int i = 0; i<this->dimension; i++) {
            for (int j = i + 1; j<this->dimension; j++) {
                norm += 2.0 * differences[i] * differences[j] * this->iSigma[i][j];
            }
            norm += differences[i] * differences[i] * this->iSigma[i][i];
        }
    } else {
        norm = 0;
        for (int i = 0; i < this->dimension; i++) {
            norm += differences[i] * differences[i] * this->iSigmaDiag[i];
        }
    }
    logDensity = common - norm * 0.5;

    /*desallocation memoire*/
    delete[] differences;
    /*fin desallocation memoire*/
    return logDensity;
}

double Gaussian::computeLogDensityNoise(int nbPoints,double **values) {
    /*declaration de variables*/
    int i, j;
    double *valeursMinimum, *valeursMaximum, logDensiteLoiUniforme;
    /*fin declaration de variables*/

    /*desallocation memoire*/
    valeursMinimum = new double[this->dimension];
    valeursMaximum = new double[this->dimension];
    /*fin allocation memoire*/

    for (i = 0; i < this->dimension; i++) {
        valeursMinimum[i] = values[i][0];
        valeursMaximum[i] = values[i][0];
    }

    for (i = 0; i < nbPoints; i++) {
        for (j = 0; j < this->dimension; j++) {
            if (valeursMinimum[j] > values[j][i]) {
                valeursMinimum[j] = values[j][i];
            }
            if (valeursMaximum[j] < values[j][i]) {
                valeursMaximum[j] = values[j][i];
            }
        }
    }
    logDensiteLoiUniforme = 0.0;
    for (i = 0; i < this->dimension; i++) {
        logDensiteLoiUniforme -= log(3.0 * (valeursMaximum[i] - valeursMinimum[i]));
    }

    /*desallocation memoire*/
    delete []valeursMinimum;
    delete []valeursMaximum;
    /*fin desallocation memoire*/

    return logDensiteLoiUniforme;
}

void Gaussian::print() {
    cout << "mu" << endl;
    for (int i = 0; i<this->dimension; i++) {
        cout << this->mu[i] << endl;
    }

    cout << "inverse sigma" << endl;
    if (diag == false) {
        for (int i = 0; i<this->dimension; i++) {
            for (int j = 0; j<this->dimension; j++) {
                cout << this->iSigma[i][j] << " ";
            }
            cout << endl;
        }
    } else {
        for (int j = 0; j<this->dimension; j++) {
            cout << this->iSigmaDiag[j] << endl;
        }
    }
}

int Gaussian::complexity() {
    int temp;
    if (diag == true) {
        temp = 2 * dimension;
    } else {
        temp = dimension + (dimension * (dimension + 1)) / 2;
    }
    return temp;
}
