/* 
 * File:   pointsSet.cpp
 * Author: wicker
 * 
 * Created on 8 juillet 2011, 14:14
 */
#include <iostream>
#include <fstream>
#include <sstream>
#include <cfloat>
#include <cmath>
#include <string>

#define MATHLIB_STANDALONE
#include <Rmath.h>
#include <vector>

#include "PointsSet.h"
using namespace std;

PointsSet::PointsSet() {
    this->size = 0;
    this->dimension = 0;
    this->clusters = NULL;
}

PointsSet::PointsSet(char *fileName) {
    this->clusters = NULL;
    /*declaration des variables*/
    ifstream f1(fileName, ios::in);
    string line;

    /*look for the dimension*/

    getline(f1, line);
    istringstream iss(line, istringstream::in);
    this->dimension = 0;
    string name;
    double value;
    iss >> name;
    while (iss.eof() == false) {
        iss >> value;
        this->dimension++;
    }

    /*read a first time to count the number of lines*/
    this->size = 0;
    while (f1.eof() == false) {
        getline(f1, line);
        this->size++;
    }
    f1.close();


    /*allocate memory*/
    int i, j;
    this->values = new double*[this->dimension];
    for (i = 0; i<this->dimension; i++) {
        this->values[i] = new double[this->size];
    }
    this->names = new string[this->size];

    /*read a second time*/
    f1.open(fileName, ios::in);
    for (i = 0; i<this->size; i++) {
        getline(f1, line);
        istringstream iss(line, istringstream::in);
        iss >> this->names[i];
        int j = 0;
        while (iss.eof() == false) {
            iss >> this->values[j][i];
            j++;
        }

    }
    f1.close();
}

PointsSet::PointsSet(int size, int dimension) {
    this->size = size;
    this->dimension = dimension;
    this->values = new double*[this->dimension];
    for (int i = 0; i<this->dimension; i++) {
        this->values[i] = new double[this->size];
    }
    this->clusters = NULL;
}

PointsSet::~PointsSet() {
}

void PointsSet::standardization() {
    int i, j;
    double mean, sd;
    for (i = 0; i<this->size; i++) {
        this->dimension = dimension;
        mean = 0;
        for (j = 0; j<this->dimension; j++) {
            mean += this->values[j][i];
        }
        mean /= (double) this->dimension;

        sd = 0;
        for (j = 0; j<this->dimension; j++) {
            sd += (this->values[j][i] - mean)*(this->values[j][i] - mean);
        }
        sd = sqrt(sd / (double) this->dimension);

        if (sd < DBL_MIN) {
            cout << "Standard deviation of " << this->names[i] << " is too low : " << sd << endl;
            exit(1);
        }

        for (j = 0; j<this->dimension; j++) {
            this->values[j][i] = (this->values[j][i] - mean) / sd;
        }
    }
}

void PointsSet::sphereProjection() {
    int i, j;
    double **coordinates, **newCoordinates;

    /*allocate memory*/
    coordinates = new double*[this->dimension];
    newCoordinates = new double*[this->dimension];
    for (i = 0; i<this->dimension; i++) {
        coordinates[i] = new double[this->size];
        newCoordinates[i] = new double[this->size];
    }

    /*copie des coordonnees d'individus vers coordonnees individus*/
    for (i = 0; i < this->size; i++) {
        for (j = 0; j < this->dimension; j++) {
            coordinates[j][i] = this->values[j][i];
        }
    }

    /*calcul des nouvelles coordonnees*/
    double *rotationVector, *angles;

    /*allocate memory*/
    rotationVector = new double[this->dimension];
    angles = new double[this->dimension];


    for (i = 0; i < this->dimension; i++) {
        rotationVector[i] = 1.0 / sqrt((double) this->dimension);
    }
    cartesiennes2polaires(this->dimension, rotationVector, angles);

    calculCoordonneesApresRotation(this->dimension, this->size, coordinates,
            angles, newCoordinates);

    /*on ramene les points sur une sphere de rayon 1*/
    double radius;
    for (i = 0; i < this->size; i++) {
        radius = 0;
        for (j = 0; j < this->dimension; j++) {
            radius += pow(newCoordinates[j][i], 2.0);
        }
        radius = sqrt(radius);
        for (j = 0; j < this->dimension; j++) {
            newCoordinates[j][i] /= radius;
        }
    }

    /*copie de nouvellesCoordonnees dans les coordonnees d'individus*/
    for (i = 0; i < this->dimension - 1; i++) {
        //printf("%s : ",individus[i]->nom);
        for (j = 0; j < this->size; j++) {
            this->values[i][j] = newCoordinates[i + 1][j];
        }
        //    printf("\n");
    }

    /*free memory*/
    delete []rotationVector;
    delete []angles;
    for (i = 0; i<this->dimension; i++) {
        delete []coordinates[i];
        delete []newCoordinates[i];
    }
    delete []coordinates;
    delete []newCoordinates;

    this->dimension = this->dimension - 1;
}

double PointsSet::computeSquaredDistance(int index, double *values) {
    double temp;
    double distance = 0;
    for (int i = 0; i<this->dimension; i++) {
        temp = this->values[i][index] - values[i];
        distance += temp*temp;
    }
    return distance;
}

void PointsSet::print() {
    for (int i = 0; i<this->size; i++) {
        for (int j = 0; j<this->dimension; j++) {
            cout << this->values[j][i] << " ";
        }
        if (this->clusters != NULL) {
            cout << "cluster : " << this->clusters[i];
        }
        cout << endl;
    }
}

void PointsSet::print(string fileName) {
    ofstream myfile;
    myfile.open(fileName.data());
    for (int i = 0; i<this->size; i++) {
        if (this->clusters != NULL) {
            myfile << this->names[i] << " " << this->clusters[i] << endl;
        }
    }
    myfile.close();

}

void PointsSet::printClusters(string fileName,bool noiseComponentEmpty) {
    ofstream myfile;
    myfile.open(fileName.data());

    /*count clusters*/
    int nbClusters = -1;
    for (int i = 0; i<this->size; i++) {
        if (this->clusters[i] > nbClusters) {
            nbClusters = this->clusters[i];
        }
    }
    nbClusters++;

    /*build clusters*/
    vector <int> *clusters = new vector<int>[nbClusters];
    int cluster;
    for (int i = 0; i<this->size; i++) {
        cluster = this->clusters[i];
        clusters[cluster].push_back(i);
    }

    /*write clusters*/
    if(noiseComponentEmpty==false) {
        myfile << "Number of clusters : " << (nbClusters-1) << " ; and a noise component" << endl << endl;
    } else {
        myfile << "Number of clusters : " << nbClusters << endl << endl;
    }

    for (cluster = 0; cluster < nbClusters; cluster++) {
        if((noiseComponentEmpty==false)&&(cluster==(nbClusters-1))) {
            myfile << "Noise Cluster " << cluster << " ; size=" << clusters[cluster].size() << endl;
        } else {
            myfile << "Cluster " << cluster << " ; size=" << clusters[cluster].size() << endl;
        }
        for (int i = 0; i < clusters[cluster].size(); i++) {
            myfile << this->names[clusters[cluster].at(i)] << endl;
        }
        myfile << endl;
    }

    myfile.close();
}

void PointsSet::kmeans(int nbClasses) throw (string) {
    int i, j, *sizeParClasse, *solution, compteurIterations1, compteurIterations2;
    int nbKmeans, nbIterationsMax, *pointsSelectionnes;
    double **centres, inertieInterClassePrecedente, inertieInterClasseCourante;
    double meilleureInertieInterClasse, *centreGravite;

    /*allocation memoire*/
    if (this->clusters == NULL) {
        this->clusters = new int[this->size];
    }
    centres = new double *[nbClasses];
    for (i = 0; i < nbClasses; i++) {
        centres[i] = new double[this->dimension];
    }
    sizeParClasse = new int [nbClasses];
    solution = new int[this->size];
    pointsSelectionnes = new int[this->size];
    centreGravite = new double[this->dimension];
    /*fin allocation memoire*/

    nbKmeans = 3;
    nbIterationsMax = 20;
    meilleureInertieInterClasse = NOT_DEFINED;

    /*calcul du centre de gravite de l'ensemble des points*/
    for (i = 0; i<this->dimension; i++) {
        centreGravite[i] = 0;
    }
    for (i = 0; i<this->size; i++) {
        for (j = 0; j<this->dimension; j++) {
            centreGravite[j] += this->values[j][i];
        }
    }
    for (i = 0; i<this->dimension; i++) {
        centreGravite[i] /= (double) this->size;
    }

    /*kmeans*/
    for (compteurIterations1 = 0; compteurIterations1 < nbKmeans; compteurIterations1++) {
        for (i = 0; i < nbClasses; i++) {
            pointsSelectionnes[i] = -1;
            while (pointsSelectionnes[i] == -1) {
                pointsSelectionnes[i] =
                        (int) floor((double) this->size - 1.0) * runif(0, 1);
                for (j = 0; j < i; j++) {
                    if (pointsSelectionnes[i] == pointsSelectionnes[j]) {
                        pointsSelectionnes[i] = -1;
                        break;
                    }
                }
            }

            for (j = 0; j<this->dimension; j++) {
                centres[i][j] = this->values[j][pointsSelectionnes[i]];
            }
        }
        /*on affecte les individus a leurs groupes les plus proches*/
        try {
            this->affectationIndividusKmeans(nbClasses, sizeParClasse,
                    centres, &inertieInterClassePrecedente,
                    centreGravite);
        } catch (string str) {
            throw (str);
        }

        compteurIterations2 = 0;
        while (1) {

            /*on affecte les individus a leurs groupes les plus proches*/
            try {
                this->affectationIndividusKmeans(nbClasses, sizeParClasse, centres,
                        &inertieInterClasseCourante, centreGravite);
            } catch (string str) {
                break;
            }

            if ((-inertieInterClasseCourante + inertieInterClassePrecedente < DBL_MIN) &&
                    (inertieInterClasseCourante - inertieInterClassePrecedente < DBL_MIN)) {
                break;
            } else {
                inertieInterClassePrecedente = inertieInterClasseCourante;
            }

            compteurIterations2++;

            if (compteurIterations2 > nbIterationsMax) {
                break;
            }
        }

        if ((meilleureInertieInterClasse = NOT_DEFINED) ||
                (inertieInterClasseCourante > meilleureInertieInterClasse)) {
            for (i = 0; i<this->size; i++) {
                solution[i] = this->clusters[i];
            }
            meilleureInertieInterClasse = inertieInterClasseCourante;
        }
    }
    for (i = 0; i<this->size; i++) {
        this->clusters[i] = solution[i];
    }

    /*desallocation memoire*/
    for (i = 0; i < nbClasses; i++) {
        delete []centres[i];
    }
    delete []centres;
    delete []sizeParClasse;
    delete []solution;
    delete []pointsSelectionnes;
    delete []centreGravite;
    /*fin desallocation memoire*/

}

/***************************************************/
/*                                                 */
/*Procedure d'affectation des individus aux groupes*/
/*et de calcul des nouveaux centres de gravite     */
/*                                                 */

/***************************************************/

void PointsSet::affectationIndividusKmeans(int nbClasses, int *sizeParClasse,
        double **centres, double *inertieInterClasse, double *centreGravite) throw (string) {
    /*declaration des variables*/
    int i, j, groupeLePlusProche;
    double distanceCourante, distanceMin, temp1, temp2;
    /*fin declaration des variables*/

    for (i = 0; i < nbClasses; i++) {
        sizeParClasse[i] = 0;
    }

    for (i = 0; i<this->size; i++) {
        distanceMin = this->computeSquaredDistance(i, centres[0]);
        groupeLePlusProche = 0;

        for (j = 1; j < nbClasses; j++) {
            distanceCourante = this->computeSquaredDistance(i, centres[j]);
            if (distanceCourante < distanceMin) {
                distanceMin = distanceCourante;
                groupeLePlusProche = j;
            }
        }
        this->clusters[i] = groupeLePlusProche;
        (sizeParClasse[groupeLePlusProche])++;
    }

    for (i = 0; i < nbClasses; i++) {
        for (j = 0; j<this->dimension; j++) {
            centres[i][j] = 0;
        }
    }

    for (i = 0; i<this->size; i++) {
        for (j = 0; j<this->dimension; j++) {
            centres[this->clusters[i]][j] += this->values[j][i];
        }
    }

    for (i = 0; i < nbClasses; i++) {
        for (j = 0; j<this->dimension; j++) {
            if (sizeParClasse[i] == 0) {
                throw (string("empty cluster"));
            }
            centres[i][j] /= (double) sizeParClasse[i];
        }
    }
    *inertieInterClasse = 0;
    for (i = 0; i < nbClasses; i++) {
        temp1 = 0;
        for (j = 0; j<this->dimension; j++) {
            temp2 = centres[i][j] - centreGravite[j];
            temp1 += temp2*temp2;
        }
        *inertieInterClasse += temp1 * (double) sizeParClasse[i];
    }

}

/***********************************************/
/*                                             */
/*Procedure de calcul des nouvelles coordonnees*/
/*apres rotation du repere                     */
/*                                             */

/***********************************************/

void calculCoordonneesApresRotation(int dimension, int size, double **coordinates,
        double *angles, double **coordinatesAfterRot) {
    int i, j;
    double val1, val2, *cosinus, *sinus;

    /*memory allocation*/
    cosinus = new double[dimension - 1];
    sinus = new double[dimension - 1];

   

    for (i = 0; i < dimension - 1; i++) {
        cosinus[i] = cos(angles[i]);
        sinus[i] = sin(angles[i]);
    }

    for (i = 0; i < size; i++) {
        for (j = 0; j < dimension; j++) {
            coordinatesAfterRot[j][i] = coordinates[j][i];
        }

        for (j = dimension - 2; j >= 0; j--) {
            val1 = coordinatesAfterRot[j][i];
            val2 = coordinatesAfterRot[j + 1][i];

            coordinatesAfterRot[j][i] = cosinus[j] * val1 + sinus[j] * val2;
            coordinatesAfterRot[j + 1][i] = -sinus[j] * val1 + cosinus[j] * val2;
        }
    }

    /*free memory*/
    delete []cosinus;
    delete []sinus;
}

/***********************************************/
/*                                             */
/*Procedure de calcul des nouvelles coordonnees*/
/*apres rotation du repere                     */
/*                                             */

/***********************************************/

void calculCoordonneesApresRotationPourUn(int dimension, double *coordinates,
        double *angles, double *coordinatesAfterRot) {
    int i, j;
    double val1, val2, *cosinus, *sinus;

    /*memory allocation*/
    cosinus = new double[dimension - 1];
    sinus = new double[dimension - 1];


    for (i = 0; i < dimension - 1; i++) {
        cosinus[i] = cos(angles[i]);
        sinus[i] = sin(angles[i]);
    }


    for (j = 0; j < dimension; j++) {
        coordinatesAfterRot[j] = coordinates[j];
    }

    for (j = dimension - 2; j >= 0; j--) {
        val1 = coordinatesAfterRot[j];
        val2 = coordinatesAfterRot[j + 1];

        coordinatesAfterRot[j] = cosinus[j] * val1 + sinus[j] * val2;
        coordinatesAfterRot[j + 1] = -sinus[j] * val1 + cosinus[j] * val2;
    }


    /*free memory*/
    delete []cosinus;
    delete []sinus;
}

/*********************************************************/
/*                                                       */
/*Procedure de calcul des coordonnees des points projetes*/
/*                                                       */

/*********************************************************/

void calculCoordonneesApresProjection(int nbDimensions, int nbPoints, double **pointsRot, double **pointsProjetes) {
    /*declaration de variables*/
    int i, j;
    double diviseur;
    /*fin declaration de variables*/

    for (i = 0; i < nbPoints; i++) {
    

        if (fabs(1.0 + pointsRot[0][i]) > DBL_MIN) {
            diviseur = 1.0 / (1.0 + pointsRot[0][i]);
            for (j = 0; j < nbDimensions - 1; j++) {
                pointsProjetes[j][i] = pointsRot[j + 1][i] * diviseur;
            }
        } else {
            for (j = 0; j < nbDimensions - 1; j++) {
                pointsProjetes[j][i] = DBL_MAX;
            }
        }
    }


}

/*********************************************************/
/*                                                       */
/*Procedure de calcul des coordonnees des points projetes*/
/*                                                       */

/*********************************************************/

void calculCoordonneesApresProjectionPourUn(int nbDimensions, double *pointRot, double *pointProjete) {
    /*declaration de variables*/
    int i, j;
    double diviseur;
    /*fin declaration de variables*/


    if (fabs(1.0 + pointRot[0]) > DBL_MIN) {
        diviseur = 1.0 / (1.0 + pointRot[0]);
        for (j = 0; j < nbDimensions - 1; j++) {
            pointProjete[j] = pointRot[j + 1] * diviseur;
        }
    } else {
        for (j = 0; j < nbDimensions - 1; j++) {
            pointProjete[j] = DBL_MAX;
        }
    }

}


/**********************************************/
/*                                            */
/*Procedure de calcul des coordonnees polaires*/
/*en fonction des coordonnees cartesiennes    */
/*                                            */

/**********************************************/

void cartesiennes2polaires(int dimension, double *cart, double *pol) {
    int i, j;
    double temp;

    pol[0] = acos(cart[0]);
    for (i = 1; i < dimension - 2; i++) {
        pol[i] = cart[i];
        for (j = 0; j < i; j++) {
            if (fabs(sin(pol[j])) > DBL_MIN) {
                pol[i] /= sin(pol[j]);
            }
        }
        pol[i] = acos(pol[i]);
    }

    pol[dimension - 2] = cart[dimension - 2];
    temp = cart[dimension - 1];
    for (j = 0; j < dimension - 2; j++) {
        if (fabs(sin(pol[j])) > DBL_MIN) {
            pol[dimension - 2] /= sin(pol[j]);
            temp /= sin(pol[j]);
        }
    }

    if (temp > 0) {
        pol[dimension - 2] = acos(pol[dimension - 2]);
    } else {
        pol[dimension - 2] = -acos(pol[dimension - 2]);
    }
}

/**************************************************/
/*                                                */
/*Procedure de calcul du mu moyen pour initialiser*/
/*la maximisation de la vraisemblance             */
/*                                                */

/**************************************************/

void calculMuMoyenSpherique(int nbPoints, double **coordonnees, int nbDimensions, double *mu) {
    /*declaration de variables*/
    int i, j;
    double norme;
    /*fin declaration de variables*/

    for (j = 0; j < nbDimensions; j++) {
        mu[j] = 0;
    }

    for (i = 0; i < nbPoints; i++) {
        for (j = 0; j < nbDimensions; j++) {

            mu[j] += coordonnees[j][i];
        }
    }
   norme = 0;
    for (j = 0; j < nbDimensions; j++) {
        mu[j] /= (double) nbPoints;
        norme += mu[j] * mu[j];
    }
    norme = sqrt(norme);
    for (j = 0; j < nbDimensions; j++) {
        mu[j] /= norme;
     }
}


/*************************************************************************************/
/*                                                                                   */
/*Procedure de calcul d'une sphere unitaire a nbDimensions, dans R^{nbDimensions+1}*/
/*                                                                                   */
/*************************************************************************************/

void calculSurfaceSphere(int nbDimensions,double *surfaceSphere) {
  /*declaration de variables*/
  int i;
  /*fin declaration de variables*/

  if(nbDimensions%2==0)
    {
      *surfaceSphere=2.0;
      for(i=1;i<=nbDimensions/2;i++)
	{
	  (*surfaceSphere)*=2.0*M_PI/(double)(2*i-1);
	}
    }
  else
    {
      *surfaceSphere=2.0*M_PI;
      for(i=1;i<=(nbDimensions-1)/2;i++)
	{
	  (*surfaceSphere)*=M_PI/(double)i;
	}
    }
}


/**************************************************************/
/*                                                            */
/*Procedure de calcul du volume d'un tetraedre a nb_dimensions*/
/*                                                            */
/**************************************************************/

void calculVolumeSimplex(int nbDimensions,double *volumeSimplex) {
  /*declaration de variables*/
  int i;
  /*fin declaration de variables*/

  *volumeSimplex=1.0;
  for(i=1;i<=nbDimensions;i++)
    {
      *volumeSimplex/=(double)i;
    }
  *volumeSimplex*=sqrt((double)(nbDimensions+1));
}