/* 
 * File:   pointsSet.h
 * Author: wicker
 *
 * Created on 8 juillet 2011, 14:14
 */

#include <string>
#include <stdlib.h>

#ifndef POINTSSET_H
#define	POINTSSET_H

#define NAME_SIZE 200
#define LINE_SIZE 10000
#define NOT_DEFINED -1
using namespace std;

class PointsSet {
public:
    int size;
    int dimension;
    double **values; //first index for variable, second index for point
    int *clusters;
    string *names;
    PointsSet();
    virtual ~PointsSet();
    PointsSet(char *fileName);
    PointsSet(int size,int dimension);
    void standardization();
    void sphereProjection();
    double computeSquaredDistance(int index, double *values);
    void kmeans(int nbClasses) throw (string);
    void print();
    void print(string fileName);
    void printClusters(string fileName,bool noiseComponentEmpty);
private:
    void affectationIndividusKmeans(int nbClasses, int *sizeParClasse,
            double **centres, double *inertieInterClasse,double *centreGravite) throw (string);
};

/***********************************************/
/*                                             */
/*Procedure de calcul des nouvelles coordonnees*/
/*apres rotation du repere                     */
/*                                             */
/***********************************************/

void calculCoordonneesApresRotation(int dimension, int size, double **coordinates,
        double *angles, double **coordinatesAfterRot);


/***********************************************/
/*                                             */
/*Procedure de calcul des nouvelles coordonnees*/
/*apres rotation du repere                     */
/*                                             */

/***********************************************/

void calculCoordonneesApresRotationPourUn(int dimension, double *coordinates,
        double *angles, double *coordinatesAfterRot);

/*********************************************************/
/*                                                       */
/*Procedure de calcul des coordonnees des points projetes*/
/*                                                       */
/*********************************************************/

void calculCoordonneesApresProjection(int nbDimensions,int nbPoints,double **pointsRot,double **pointsProjetes);


/*********************************************************/
/*                                                       */
/*Procedure de calcul des coordonnees des points projetes*/
/*                                                       */
/*********************************************************/

void calculCoordonneesApresProjectionPourUn(int nbDimensions,double *pointRot,double *pointProjete);

/**********************************************/
/*                                            */
/*Procedure de calcul des coordonnees polaires*/
/*en fonction des coordonnees cartesiennes    */
/*                                            */

/**********************************************/

void cartesiennes2polaires(int dimension, double *cart, double *pol);

/**************************************************/
/*                                                */
/*Procedure de calcul du mu moyen pour initialiser*/
/*la maximisation de la vraisemblance             */
/*                                                */
/**************************************************/

void calculMuMoyenSpherique(int nbPoints,double **coordonnees,int nbDimensions,double *mu);


/*************************************************************************************/
/*                                                                                   */
/*Procedure de calcul d'une sphere unitaire a nbDimensions, dans R^{nbDimensions+1}*/
/*                                                                                   */
/*************************************************************************************/

void calculSurfaceSphere(int nbDimensions,double *surfaceSphere);

/**************************************************************/
/*                                                            */
/*Procedure de calcul du volume d'un tetraedre a nb_dimensions*/
/*                                                            */
/**************************************************************/

void calculVolumeSimplex(int nbDimensions,double *volumeSimplex);
#endif	/* POINTSSET_H */

