/* 
 * File:   SGaussian.cpp
 * Author: wicker
 * 
 * Created on 5 septembre 2011, 14:29
 */

#include <iostream>
#include <cfloat>
#include <cmath>
#include "Lawc.h"
#include "SGaussian.h"
#include "matrix.h"
#include "PointsSet.h"
#define MATHLIB_STANDALONE
#include <Rmath.h>
using namespace std;

SGaussian::SGaussian() {
}

SGaussian::SGaussian(int dimension, bool diag) {

    this->dimension = dimension;
    this->diag = false;

    this->mu = new double[this->dimension];

    this->iSigma = new double *[this->dimension];
    for (int i = 0; i<this->dimension; i++) {
        this->iSigma[i] = new double[this->dimension];
    }

}

SGaussian::~SGaussian() {
}

void SGaussian::estimation(int nbPoints, double **values) {
    /*declaration de variables*/
    int i, j, k, l, nbMus, select1, select2, nbEtapes, *ordreMus, tailleSelection;
    double *muMoyen, *muCourant, *meilleurMu, meilleureLogVraisemblance, logVraisemblanceCourante;
    double **pointsRot, **pointsProjetes, *angles, **matriceCovariance, normeCarre;
    double **muPopulation, *logVrais, **meilleursMus, commun, **meilleurISigma;
    double meilleurLogDeterminant;
    /*fin declaration de variables*/

    meilleureLogVraisemblance = -DBL_MAX;
    nbEtapes = 5;
    nbMus = 1;
    tailleSelection = 5;

    /*allocation memoire*/
    pointsProjetes = new double *[this->dimension];
    for (i = 0; i < this->dimension; i++) {
        pointsProjetes[i] = new double[nbPoints];
    }
    pointsRot = new double*[this->dimension];
    for (i = 0; i < this->dimension; i++) {
        pointsRot[i] = new double[nbPoints];
    }
    angles = new double[this->dimension - 1];
    meilleurMu = new double[this->dimension];
    meilleurISigma = new double *[this->dimension - 1];
    for (i = 0; i < this->dimension - 1; i++) {

        meilleurISigma[i] = new double[this->dimension - 1];
    }
    /*fin allocation memoire*/

    int compteurMus, pointSelectionne;

    /*calcul du mu moyen*/
    for (compteurMus = 0; compteurMus < nbMus; compteurMus++) {
        if (compteurMus == 0) {
            calculMuMoyenSpherique(nbPoints, values, this->dimension, this->mu);
        } else {
            pointSelectionne = (int) floor(runif(0, (double) nbPoints));
            for (i = 0; i < this->dimension; i++) {
                this->mu[i] = values[i][pointSelectionne];
            }
        }

        /*recuperation des angles de mu*/
        cartesiennes2polaires(this->dimension, this->mu, angles);


        /*calcul des nouvelles coordonnees apres rotation du repere*/
        calculCoordonneesApresRotation(this->dimension, nbPoints, values, angles,
                pointsRot);


        /*calcul des coordonnees des points projetes*/
        calculCoordonneesApresProjection(this->dimension, nbPoints, pointsRot,
                pointsProjetes);

        /*estimation de la matrice de variance-covariance*/
        double temp;

        for (i = 0; i<this->dimension - 1; i++) {
            for (j = i; j<this->dimension - 1; j++) {
                this->iSigma[i][j] = 0;
                for (k = 0; k < nbPoints; k++) {
                    this->iSigma[i][j] += pointsProjetes[i][k] * pointsProjetes[j][k];
                }
                this->iSigma[i][j] /= (double) (nbPoints - 1);
                this->iSigma[j][i] = this->iSigma[i][j];
            }
        }
        /*on inverse la matrice de covariance et on recupere au passage*/
        /*le log du determinant de cette meme matrice*/
        this->logDeterminant = inversionMatrix(dimension - 1, this->iSigma);

        logVraisemblanceCourante = 0;
        commun = -this->logDeterminant * 0.5 - (double) (dimension - 1)*0.5 * log(2.0 * M_PI);

        /*calcul des vraisemblances*/
        for (j = 0; j < nbPoints; j++) {
            normeCarre = 0;
            for (k = 0; k<this->dimension; k++) {
                for (l = k + 1; l<this->dimension; l++) {
                    normeCarre += 2.0 * pointsProjetes[k][j] * pointsProjetes[l][j] * this->iSigma[k][l];
                }
                normeCarre += pointsProjetes[k][j] * pointsProjetes[k][j] * this->iSigma[k][k];
            }
            // calculNorme(dimension - 1, pointsProjetes[j], this->iSigma);

            logVraisemblanceCourante -= normeCarre * 0.5;
            logVraisemblanceCourante -= (double) (dimension - 1) * log(fabs(1.0 + pointsRot[0][j]));
            logVraisemblanceCourante += commun;

        }

        if ((compteurMus == 0) || (logVraisemblanceCourante > meilleureLogVraisemblance)) {
            meilleureLogVraisemblance = logVraisemblanceCourante;
            for (i = 0; i < dimension; i++) {
                meilleurMu[i] = this->mu[i];
            }
            for (i = 0; i < dimension - 1; i++) {
                for (j = 0; j < dimension - 1; j++) {
                    meilleurISigma[i][j] = this->iSigma[i][j];
                }
            }
            meilleurLogDeterminant = this->logDeterminant;
        }
    }

    for (i = 0; i < dimension; i++) {
        this->mu[i] = meilleurMu[i];
    }
    for (i = 0; i < dimension - 1; i++) {
        for (j = 0; j < dimension - 1; j++) {
            this->iSigma[i][j] = meilleurISigma[i][j];
        }
    }
    this->logDeterminant = meilleurLogDeterminant;

    /*cout << "determinant " << this->logDeterminant << endl;
     for (i = 0; i < dimension - 1; i++) {
        for (j = 0; j < dimension - 1; j++) {
            cout << this->iSigma[i][j] << " ";
        }
        cout << endl;
    }
    int pause;

    std::cin >> pause;*/

    /*desallocation memoire*/
    for (i = 0; i < this->dimension; i++) {
        delete []pointsProjetes[i];
    }
    delete []pointsProjetes;
    for (i = 0; i < this->dimension; i++) {
        delete []pointsRot[i];
    }
    delete []pointsRot;
    delete []angles;
    delete []meilleurMu;
    for (i = 0; i < dimension - 1; i++) {
        delete []meilleurISigma[i];
    }
    delete []meilleurISigma;
    /*fin desallocation memoire*/

}

/*compute log density for one point*/

double SGaussian::computeLogDensity(double *values) {
    /*declaration de variables*/
    int i, j, k, l;
    double *pointRot, *pointProjete, somme, norme, *angles, logDensity;
    double commun;
    /*fin declaration de variables*/

    /*time_t temps_debut,temps_fin,temps_debut2,temps_fin2,temps_coordonnees;
      temps_debut=time(NULL);*/



    /*allocation memoire*/

    pointRot = new double[this->dimension];
    pointProjete = new double[this->dimension];

    angles = new double[this->dimension];
    /*fin allocation memoire*/

    /*recuperation des angles de mu*/
    cartesiennes2polaires(this->dimension, this->mu, angles);

    /*temps_debut2=time(NULL);*/

    /*calcul des nouvelles coordonnees apres rotation du repere*/
    calculCoordonneesApresRotationPourUn(this->dimension, values, angles, pointRot);

    /*calcul des coordonnees des points projetes*/
    calculCoordonneesApresProjectionPourUn(this->dimension, pointRot, pointProjete);
    /*temps_fin2=time(NULL);
      temps_coordonnees+=temps_fin2-temps_debut2;*/

    commun = -this->logDeterminant * 0.5 - (double) (this->dimension - 1)*0.5 * log(2.0 * M_PI);

    /*calcul des vraisemblances*/
    norme = calculNorme(this->dimension - 1, pointProjete, this->iSigma);
    logDensity = -norme * 0.5;
    logDensity -= (double) (this->dimension - 1) * log(fabs(1.0 + pointRot[0]));
    logDensity += commun;

    /*free memory*/
    delete []pointRot;
    delete []pointProjete;
    delete []angles;

    return logDensity;
}

double SGaussian::computeLogDensityNoise(int nbPoints, double **values) {
    double surfaceSphere, logDensity;
    calculSurfaceSphere(this->dimension - 1, &surfaceSphere);

    logDensity = -log(surfaceSphere);
    return logDensity;
}

void SGaussian::print() {
    cout << "mu" << endl;
    for (int i = 0; i<this->dimension; i++) {
        cout << this->mu[i] << endl;
    }

    cout << "inverse sigma" << endl;
    for (int i = 0; i<this->dimension - 1; i++) {
        for (int j = 0; j<this->dimension - 1; j++) {
            cout << this->iSigma[i][j] << " ";
        }
        cout << endl;
    }

}

int SGaussian::complexity() {
    int temp;
    /*a completer*/
    temp = this->dimension + (this->dimension * (this->dimension + 1)) / 2-1;

    return temp;
}
