/* 
 * File:   SStudent.cpp
 * Author: wicker
 * 
 * Created on 8 septembre 2011, 09:17
 */

#include <iostream>
#include <cfloat>
#include <cmath>
#include "Lawc.h"
#include "SStudent.h"
#include "matrix.h"
#include "PointsSet.h"
#define MATHLIB_STANDALONE
#include <Rmath.h>
using namespace std;

SStudent::SStudent() {
}

SStudent::SStudent(int dimension, bool diag) {
    this->dimension = dimension;
    this->diag = false;

    this->mu = new double[this->dimension];

    this->iSigma = new double *[this->dimension];
    for (int i = 0; i<this->dimension; i++) {
        this->iSigma[i] = new double[this->dimension];
    }

}

SStudent::~SStudent() {
}

void SStudent::estimation(int nbPoints, double **values) {
    /*declaration de variables*/
    int i, j, k, l, nbMus, compteurMus, pointSelectionne;
    double meilleurNu, *meilleurMu, meilleureLogVraisemblance, logVraisemblanceCourante;
    double **meilleurISigma, meilleurLogDeterminant;
    /*fin declaration de variables*/

    meilleureLogVraisemblance = -DBL_MAX;

    /*memory allocation*/
    meilleurMu = new double[this->dimension];
    meilleurISigma = new double*[this->dimension - 1];
    for (i = 0; i<this->dimension - 1; i++) {
        meilleurISigma[i] = new double[this->dimension - 1];
    }
    /*end of memory allocation*/

    nbMus = 1;
    for (compteurMus = 0; compteurMus < nbMus; compteurMus++) {
        /*calcul du mu moyen*/
        if (compteurMus == 0) {
            calculMuMoyenSpherique(nbPoints, values, this->dimension, this->mu);
        } else {
            pointSelectionne = (int) floor(runif(0, (double) nbPoints));
            for (i = 0; i < this->dimension; i++) {
                this->mu[i] = values[i][pointSelectionne];
            }
        }
        logVraisemblanceCourante = this->estimationAvecMuFixe(nbPoints, values);

        if ((compteurMus == 0) || (logVraisemblanceCourante > meilleureLogVraisemblance)) {
            meilleureLogVraisemblance = logVraisemblanceCourante;
            meilleurNu = this->nu;
            for (i = 0; i < this->dimension; i++) {
                meilleurMu[i] = this->mu[i];
            }
            for (i = 0; i < this->dimension - 1; i++) {
                for (j = 0; j < this->dimension - 1; j++) {
                    meilleurISigma[i][j] = this->iSigma[i][j];
                }
            }
        }
    }

    this->nu = meilleurNu;
    for (i = 0; i < this->dimension; i++) {
        this->mu[i] = meilleurMu[i];
    }
    for (i = 0; i < this->dimension - 1; i++) {
        for (j = 0; j < this->dimension - 1; j++) {
            this->iSigma[i][j] = meilleurISigma[i][j];
        }
    }

    /*desallocation memoire*/
    delete []meilleurMu;
    for (i = 0; i < dimension - 1; i++) {
        delete []meilleurISigma[i];
    }
    delete []meilleurISigma;
    /*fin desallocation memoire*/

}


/************************************************************/
/*                                                          */
/*Procedure de maximisation de la vraisemblance avec mu fixe*/
/*                                                          */

/************************************************************/

double SStudent::estimationAvecMuFixe(int nbPoints, double **values) {
    /*declaration de variables*/
    int i, j, k, l, nbMus, select1, select2, nbEtapes, *ordreMus, tailleSelection;
    double *muMoyen, *muCourant, *meilleurMu, **meilleurSigma, logVraiMax, logVraiCour;
    double **pointsRot, **pointsProjetes, *angles, normeCarre;
    double **muPopulation, *logVrais, **meilleursMus;
    /*fin declaration de variables*/

    nbEtapes = 5;
    nbMus = 1;
    tailleSelection = 5;

    /*allocation memoire*/
    pointsProjetes = new double*[this->dimension];
    for (i = 0; i < this->dimension; i++) {
        pointsProjetes[i] = new double[nbPoints];
    }
    pointsRot = new double*[this->dimension];
    for (i = 0; i < this->dimension; i++) {
        pointsRot[i] = new double[nbPoints];
    }
    angles = new double[this->dimension - 1];
    /*fin allocation memoire*/

    /*recuperation des angles de mu*/
    cartesiennes2polaires(this->dimension, this->mu, angles);

    /*calcul des nouvelles coordonnees apres rotation du repere*/
    calculCoordonneesApresRotation(this->dimension, nbPoints, values, angles, pointsRot);

    /*calcul des coordonnees des points projetes*/
    calculCoordonneesApresProjection(this->dimension, nbPoints, pointsRot, pointsProjetes);

    int dureeEM, compteur;
    double nu, logVraisemblanceCourante, meilleureLogVraisemblance;
    double commun, *taus, *meilleursTaus, sommeTaus, sommeLogTaus;

    dureeEM = 10;

    /*memory allocation*/
    taus = new double[nbPoints];
    meilleursTaus = new double[nbPoints];
    /*end of memory allocation*/

    /*initialisation des taus*/
    sommeTaus = 0;
    sommeLogTaus = 0;
    for (i = 0; i < nbPoints; i++) {
        taus[i] = 2.0;
        sommeTaus += taus[i];
        sommeLogTaus += log(taus[i]);
    }

    for (compteur = 0; compteur < dureeEM; compteur++) {
        /*Etape M*/
        /*estimation de la matrice de covariance*/
        computeCovarianceMatrixWithTaus(this->dimension - 1, nbPoints, pointsProjetes, this->iSigma, taus);

        /*estimation de nu*/
        nu = 1.0 / (-1.0 - sommeLogTaus / (double) nbPoints + sommeTaus / (double) nbPoints);

        /*calcul de la vraisemblance*/
        logVraisemblanceCourante = 0;

        /*on inverse la matrice de covariance et on recupere au passage*/
        /*le log du determinant de cette meme matrice*/
        this->logDeterminant = inversionMatrix(this->dimension - 1, this->iSigma);

        commun = lgammafn((double) (nu + this->dimension - 1)*0.5) -
                lgammafn((double) nu * 0.5)-(double) (this->dimension - 1)*0.5 * log((double) nu * M_PI) -
                this->logDeterminant * 0.5;

        /*calcul des vraisemblances*/
        for (j = 0; j < nbPoints; j++) {
            normeCarre = 0;
            for (k = 0; k<this->dimension; k++) {
                for (l = k + 1; l<this->dimension; l++) {
                    normeCarre += 2.0 * pointsProjetes[k][j] * pointsProjetes[l][j] * this->iSigma[k][l];
                }
                normeCarre += pointsProjetes[k][j] * pointsProjetes[k][j] * this->iSigma[k][k];
            }
            logVraisemblanceCourante -= (double) (nu + this->dimension - 1)*0.5 * log(1.0 + normeCarre / (double) nu);
            logVraisemblanceCourante -= (double) (this->dimension - 1) * log(fabs(1.0 + pointsRot[0][j]));
            logVraisemblanceCourante += commun;
        }

        if ((compteur == 0) || (logVraisemblanceCourante > meilleureLogVraisemblance)) {
            meilleureLogVraisemblance = logVraisemblanceCourante;
            for (i = 0; i < nbPoints; i++) {
                meilleursTaus[i] = taus[i];
            }
        }

        /*Etape E*/
        double distance;
        for (i = 0; i < nbPoints; i++) {
            distance = 0;
            for (j = 0; j < this->dimension - 1; j++) {
                for (k = 0; k < this->dimension - 1; k++) {
                    distance += pointsProjetes[j][i] * pointsProjetes[k][i] * this->iSigma[j][k];
                }
            }
            taus[i] = (nu + (double) (this->dimension - 1)) / (nu + distance);
        }
    }

    sommeTaus = 0;
    sommeLogTaus = 0;
    for (i = 0; i < nbPoints; i++) {
        sommeTaus += meilleursTaus[i];
        sommeLogTaus += log(meilleursTaus[i]);
    }
    /*estimation de la matrice de covariance*/
    computeCovarianceMatrixWithTaus(this->dimension - 1, nbPoints, pointsProjetes, this->iSigma, meilleursTaus);

    /*on inverse la matrice de covariance et on recupere au passage*/
    /*le log du determinant de cette meme matrice*/
    this->logDeterminant = inversionMatrix(this->dimension - 1, this->iSigma);

    /*estimation de nu*/
    nu = 1.0 / (-1.0 - sommeLogTaus / (double) nbPoints + sommeTaus / (double) nbPoints);
    this->nu = nu;


    /*desallocation memoire*/
    delete []taus;
    delete []meilleursTaus;
    for (i = 0; i < this->dimension; i++) {
        delete []pointsProjetes[i];
    }
    delete []pointsProjetes;
    for (i = 0; i < this->dimension; i++) {
        delete []pointsRot[i];
    }
    delete []pointsRot;
    delete []angles;
    /*fin desallocation memoire*/

    return meilleureLogVraisemblance;
}

/***********************************/
/*compute log density for one point*/

/***********************************/

double SStudent::computeLogDensity(double *values) {
    /*declaration de variables*/
    int i, j, k, l;
    double *pointRot, *pointProjete, somme, norme, *angles, logDensity;
    double commun;
    /*fin declaration de variables*/

    /*time_t temps_debut,temps_fin,temps_debut2,temps_fin2,temps_coordonnees;
      temps_debut=time(NULL);*/

    /*allocation memoire*/
    pointRot = new double[this->dimension];
    pointProjete = new double[this->dimension];

    angles = new double[this->dimension];
    /*fin allocation memoire*/

    /*recuperation des angles de mu*/
    cartesiennes2polaires(this->dimension, this->mu, angles);

    /*temps_debut2=time(NULL);*/

    /*calcul des nouvelles coordonnees apres rotation du repere*/
    calculCoordonneesApresRotationPourUn(this->dimension, values, angles, pointRot);

    /*calcul des coordonnees des points projetes*/
    calculCoordonneesApresProjectionPourUn(this->dimension, pointRot, pointProjete);

    commun = lgammafn((double) (this->nu + this->dimension - 1)*0.5) -
            lgammafn((double) this->nu * 0.5)-(double) (this->dimension - 1)*0.5 * log((double) this->nu * M_PI) -
            this->logDeterminant * 0.5;

    norme = calculNorme(this->dimension, pointProjete, this->iSigma);
    logDensity = -(double) (this->nu + this->dimension - 1)*0.5 * log(1.0 + norme / (double) this->nu);
    logDensity -= (double) (this->dimension - 1) * log(fabs(1.0 + pointRot[0]));
    logDensity += commun;

    /*fin desallocation memoire*/
    return logDensity;
}

double SStudent::computeLogDensityNoise(int nbPoints, double **values) {
    double surfaceSphere, logDensity;
    calculSurfaceSphere(this->dimension - 1, &surfaceSphere);

    logDensity = -log(surfaceSphere);
    return logDensity;
}

void SStudent::print() {
    cout << "nu " << this->nu << endl;
    cout << "mu" << endl;
    for (int i = 0; i<this->dimension; i++) {
        cout << this->mu[i] << endl;
    }

    cout << "inverse sigma" << endl;
    for (int i = 0; i<this->dimension - 1; i++) {
        for (int j = 0; j<this->dimension - 1; j++) {
            cout << this->iSigma[i][j] << " ";
        }
        cout << endl;
    }

}

int SStudent::complexity() {
    int temp;
    /*a completer*/

    temp = this->dimension + (this->dimension * (this->dimension + 1)) / 2;
    return temp;
}


