/* 
 * File:   main.cpp
 * Author: wicker
 *
 * Created on 8 juillet 2011, 10:38
 */

#include <cstdlib>
#include <string>
#include <string.h>

#include "PointsSet.h"
#include "MM.h"
#include "Gaussian.h"
#include "SGaussian.h"
#include "SStudent.h"
#include "FisherVonMises.h"
#include "Dirichlet.h"

using namespace std;

/*
 * 
 */
int main(int argc, char** argv) {
    if (argc < 4) {
        cout << "Usage : mmclustering filein fileout aic|bic [-law=gaussian|sgaussian|sstudent|fishervonmises|dirichlet] "
                << "[-nbmax=maxNumberOfClusters] [-standardization] [-nonoise]" << endl;
        exit(0);
    }
    PointsSet *myPoints = new PointsSet(argv[1]);
    string fileOut(argv[2]);
    string criterium(argv[3]);
    string lawType = "gaussian";
    bool standardization = false;
    bool noiseComponent=true;

    int maxNbComponents=-1;
    double proba = 0.01;
    char *buf;
    for (int i = 4; i < argc; i++) {
        if (strstr(argv[i], "-nbmax=") != NULL) {
            buf = strstr(argv[i], "=") + 1;
            maxNbComponents = atoi(buf);
        } else if (strstr(argv[i], "-proba=") != NULL) {
            buf = strstr(argv[i], "=") + 1;
            proba = atof(buf);
        } else if (strstr(argv[i], "-standardization") != NULL) {
            standardization = true;
        } else if (strstr(argv[i], "-nonoise") != NULL) {
            noiseComponent = false;
        } else if (strstr(argv[i], "-law=") != NULL) {
            buf = strstr(argv[i], "=") + 1;
            lawType = buf;
        }
    }
    if (standardization == true) {
        myPoints->standardization();
        myPoints->sphereProjection();
    }

    bool noiseComponentEmpty;
    if (lawType.compare("gaussian") == 0) {
        MM <Gaussian>*monMelange = new MM<Gaussian > (criterium, myPoints, proba, maxNbComponents,noiseComponent);
        monMelange->expectation(myPoints);
        noiseComponentEmpty=monMelange->noiseComponentEmpty;
    } else if (lawType.compare("sgaussian")==0) {
        MM <SGaussian>*monMelange = new MM<SGaussian > (criterium, myPoints, proba, maxNbComponents,noiseComponent);
        monMelange->expectation(myPoints);
        noiseComponentEmpty=monMelange->noiseComponentEmpty;
    } else if (lawType.compare("sstudent")==0) {
        MM <SStudent>*monMelange = new MM<SStudent > (criterium, myPoints, proba, maxNbComponents,noiseComponent);
        monMelange->expectation(myPoints);
        noiseComponentEmpty=monMelange->noiseComponentEmpty;
    } else if (lawType.compare("fishervonmises")==0) {
        MM <FisherVonMises>*monMelange = new MM<FisherVonMises > (criterium, myPoints, proba, maxNbComponents,noiseComponent);
        monMelange->expectation(myPoints);
        noiseComponentEmpty=monMelange->noiseComponentEmpty;
    } else if (lawType.compare("dirichlet")==0) {
        MM <Dirichlet>*monMelange = new MM<Dirichlet > (criterium, myPoints, proba, maxNbComponents,noiseComponent);
        monMelange->expectation(myPoints);
        noiseComponentEmpty=monMelange->noiseComponentEmpty;
    }

        myPoints->printClusters(fileOut,noiseComponentEmpty);



    return 0;
}


