#include <cmath>
#include <cfloat>

#include "matrix.h"



/*******************************************************************/
/*                                                                 */
/*Procedure d'inversion de matrice qui renvoie le log du determinant*/
/*de la matrice a inverser ou 0 s'il y a un depassement de capacite*/
/*Nota bene : matrice est modifiee par la procedure                */
/*                                                                 */
/*******************************************************************/

double inversionMatrix(int dimension,double **matrix)
{
  /*declaration de variables*/
  int i,j,k,pivot;
  double **inverseMatrix,value;
  double determinant;
  /*fin declaration de variables*/

  /*allocation memoire*/
  inverseMatrix=new double *[dimension];
  for(i=0;i<dimension;i++) {
      inverseMatrix[i]=new double[dimension];
    }
  /*fin allocation memoire*/

  for(i=0;i<dimension;i++) {
      for(j=0;j<dimension;j++)
	{
	  inverseMatrix[i][j]=0;
	}
      inverseMatrix[i][i]=1.0;
    }

  determinant=0.0;
  for(i=0;i<dimension;i++)
    {
      pivot=-1;
      for(j=i;j<dimension;j++)
	{
	  if(fabs(matrix[j][i])>DBL_MIN)
	    {
	      pivot=j;
	      break;
	    }
	}

      if(pivot==-1)
	{
	  return 0.0;
	}
      echangeLignes(dimension,matrix,i,pivot);
      echangeLignes(dimension,inverseMatrix,i,pivot);

      value=matrix[i][i];

      for(j=0;j<dimension;j++)
	{
	  matrix[i][j]/=value;
	  inverseMatrix[i][j]/=value;
	}
      determinant+=log(fabs(value));

      for(j=0;j<dimension;j++)
	{
	  if(j!=i)
	    {
	      value=matrix[j][i];
	      for(k=0;k<dimension;k++)
		{
		  matrix[j][k]-=value*matrix[i][k];
		  inverseMatrix[j][k]-=value*inverseMatrix[i][k];
		}
	    }
	}

    }

  for(i=0;i<dimension;i++)
    {
      for(j=0;j<dimension;j++)
	{
	  matrix[i][j]=inverseMatrix[i][j];
	}
    }

  /*desallocation memoire*/
  for(i=0;i<dimension;i++) {
      delete []inverseMatrix[i];
  }
  delete []inverseMatrix;
  /*fin desallocation memoire*/

  return determinant;
}

/*****************************************************/
/*                                                   */
/*Procedure d'echange de deux lignes dans une matrice*/
/*                                                   */
/*****************************************************/

void echangeLignes(int dimension,double **matrix,int line1,int line2) {
  /*declaration de variables*/
  int i;
  double valTemp;
  /*fin declaration de variables*/

  for(i=0;i<dimension;i++) {
      valTemp=matrix[line1][i];
      matrix[line1][i]=matrix[line2][i];
      matrix[line2][i]=valTemp;
    }
}


/*******************************************************************************/
/*                                                                             */
/*Procedure de calcul de la norme au carre d'un vecteur pour une matrice donnee*/
/*                                                                             */
/*******************************************************************************/

double calculNorme(int nbDimensions,double *vecteur,double **matrice) {
  /*declaration de variables*/
  int i,j;
  double normeCarre;
  /*fin declaration de variables*/

  /*calcul du produit de gauche*/
  normeCarre=0;
  for(i=0;i<nbDimensions;i++)
    {
      for(j=i+1;j<nbDimensions;j++)
	{
	  normeCarre+=2.0*vecteur[i]*vecteur[j]*matrice[j][i];
	}
      normeCarre+=vecteur[i]*vecteur[i]*matrice[i][i];
    }

  return normeCarre;
}

/***************************************************/
/*                                                 */
/*Procedure to compute a variance-covariance matrix*/
/*                                                 */
/***************************************************/

void computeCovarianceMatrix(int dimension,int nbPoints,double **values,double **sigma) {
    int i,j,k;
    for (i = 0; i<dimension - 1; i++) {
        for (j = i; j<dimension - 1; j++) {
            sigma[i][j] = 0;
            for (k = 0; k < nbPoints; k++) {
                sigma[i][j] += values[i][k] * values[j][k];
            }
            sigma[i][j] /= (double) (nbPoints - 1);
            sigma[j][i] = sigma[i][j];
        }
    }
}

/************************************************************************/
/*                                                                      */
/*Procedure to compute a variance-covariance matrix with parameters taus*/
/*                                                                      */
/************************************************************************/

void computeCovarianceMatrixWithTaus(int dimension,int nbPoints,double **values,double **sigma,double *taus) {
        int i,j,k;
    for (i = 0; i<dimension - 1; i++) {
        for (j = i; j<dimension - 1; j++) {
            sigma[i][j] = 0;
            for (k = 0; k < nbPoints; k++) {
                sigma[i][j] += taus[k]*values[i][k] * values[j][k];
            }
            sigma[i][j] /= (double) (nbPoints - 1);
            sigma[j][i] = sigma[i][j];
        }
    }
}